import { Store } from "@willsoto/node-konfig-core";
import { retry, type RetryOptions } from "jsr:@std/async";

type FetchInputParameter = Parameters<typeof fetch>[0];
type FetchInitParameterWithClient =
    | RequestInit
    | RequestInit & { client: Deno.HttpClient };
type FetchReturn = ReturnType<typeof fetch>;

let proxies: string[] = [];
let currentProxyIndex = 0;
let proxyData: string;
try {
    proxyData = Deno.readTextFileSync("proxies.txt");
    proxies = proxyData.split("\n").map((line) => line.trim()).filter((line) =>
        line.length > 0
    );
    console.log("[INFO] Num of proxies on the list:", proxies.length)
} catch (error) {
    console.error("[ERROR] Error reading proxy file:", error);
    console.log("[INFO] Proxies from a list of proxies will not be used");
}

export const getFetchClient = (konfigStore: Store): {
    (
        input: FetchInputParameter,
        init?: FetchInitParameterWithClient,
    ): FetchReturn;
} => {
    if (
        Deno.env.get("PROXY") || konfigStore.get("networking.proxy") ||
        (proxies.length > 0)
    ) {
        return async (
            input: FetchInputParameter,
            init?: RequestInit,
        ) => {
            const proxyUrl = Deno.env.get("PROXY") ||
                konfigStore.get("networking.proxy") as string ||
                (proxies.length > 0 ? proxies[currentProxyIndex] : null);
            if (!proxyUrl) {
                throw new Error("No proxy available");
            }
            const client = Deno.createHttpClient({
                proxy: {
                    url: proxyUrl,
                },
            });

            currentProxyIndex = (currentProxyIndex + 1) % proxies.length;

            const fetchRes = await fetch(input, {
                client,
                headers: init?.headers,
                method: init?.method,
                body: init?.body,
            });

            return new Response(fetchRes.body, {
                status: fetchRes.status,
                headers: fetchRes.headers,
            });
        };
    }

    return (input: FetchInputParameter, init?: FetchInitParameterWithClient) =>
        fetchShim(konfigStore, input, init);
};

function fetchShim(
    konfigStore: Store,
    input: FetchInputParameter,
    init?: FetchInitParameterWithClient,
): FetchReturn {
    const fetchTimeout = konfigStore.get("networking.fetch_timeout_ms");
    const fetchRetry = konfigStore.get("networking.fetch_retry_enable");
    const fetchMaxAttempts = konfigStore.get("networking.fetch_retry_times");
    const fetchInitialDebounce = konfigStore.get(
        "networking.fetch_retry_initial_debounce",
    );
    const fetchDebounceMultiplier = konfigStore.get(
        "networking.fetch_retry_debounce_multiplier",
    );
    const retryOptions: RetryOptions = {
        maxAttempts: Number(fetchMaxAttempts) || 1,
        minTimeout: Number(fetchInitialDebounce) || 0,
        multiplier: Number(fetchDebounceMultiplier) || 0,
        jitter: 0,
    };

    const callFetch = () =>
        fetch(input, {
            // only set the AbortSignal if the timeout is supplied in the config
            signal: fetchTimeout
                ? AbortSignal.timeout(Number(fetchTimeout))
                : null,
            ...(init || {}),
        });
    // if retry enabled, call retry with the fetch shim, otherwise pass the fetch shim back directly
    return fetchRetry ? retry(callFetch, retryOptions) : callFetch();
}
